#Copyright 2008-2009  Marius Muja (mariusm@cs.ubc.ca). All rights reserved.
#Copyright 2008-2009  David G. Lowe (lowe@cs.ubc.ca). All rights reserved.
#
#THE BSD LICENSE
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions
#are met:
#
#1. Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#2. Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



from rp.libs.pyflann.exceptions import FLANNException
from . import binary_dataset
from . import dat_dataset
from . import npy_dataset
from . import hdf5_dataset

import os.path
from numpy import float32

dataset_formats = { 
    'bin' : binary_dataset, 
    'dat' : dat_dataset, 
    'npy' : npy_dataset,
    'hdf5' : hdf5_dataset 
}


def load(filename, rows = -1, cols = -1, dtype = float32, **kwargs):
    
    for format in list(dataset_formats.values()):
        if format.check(filename):
            return format.load(filename, rows, cols, dtype, **kwargs)
    raise FLANNException("Error: Unknown dataset format")
    
    
def save(dataset, filename, format = None, **kwargs):    
    try:
        if format is None:
            basename,extension = os.path.splitext(filename)
            format = extension[1:]
        handler = dataset_formats[format]
        handler.save(dataset, filename, **kwargs)
    except Exception as e:
        raise FLANNException(e)



from typing import Any, Optional, Union, Text

from Crypto.PublicKey.RSA import _RSAobj

class PKCS1OAEP_Cipher:
    def __init__(self, key: _RSAobj, hashAlgo: Any, mgfunc: Any, label: Any) -> None: ...
    def can_encrypt(self): ...
    def can_decrypt(self): ...
    def encrypt(self, message: Union[bytes, Text]) -> bytes: ...
    def decrypt(self, ct: bytes) -> bytes: ...


def new(key: _RSAobj, hashAlgo: Optional[Any] = ..., mgfunc: Optional[Any] = ..., label: Any = ...) -> PKCS1OAEP_Cipher: ...



from overrides import overrides

from ..json_widget import JSONWidget


class ArrayUnexpandedWidget(JSONWidget):
    """
    a widget display JSON `array` type node
    """

    def __init__(self,
                 node,
                 display_key,
                 ):
        super().__init__(node, True, display_key)

    @overrides
    def load_value_markup(self):
        return "[\u2026]"



from abc import ABCMeta
from abc import abstractmethod


class JSONCompositeEndNode(metaclass=ABCMeta):
    """
    base implementation for a end node attached to each composite node
    which represents the end of a composite node.

    this node is used to distinguish between start node, so when visiting
    each node in a tree, the widget for that position could be uniquely
    identified. this is required by JSONListWalker.

    all of the methods other than loading widget will be delegate to the
    attaching composite node.
    """

    def __init__(self, start_node):
        self._start_node = start_node
        # ui
        self._widget = None

    def is_end_node(self):
        return True

    def is_expanded(self):
        return self._start_node.is_expanded()

    def collapse_all(self):
        return self._start_node.collapse_all()

    def toggle_expanded(self):
        self._start_node.toggle_expanded()

    def get_start_node(self):
        return self._start_node

    def get_depth(self):
        return self._start_node.get_depth()

    def get_parent(self):
        return self._start_node.get_parent()

    def get_last_child(self):
        return self._start_node.get_last_child()

    # =================================================================================== #
    # ui                                                                                  #
    # =================================================================================== #
    def get_widget(self):
        if self._widget is None:
            self._widget = self.load_widget()
        return self._widget

    @abstractmethod
    def load_widget(self):
        raise NotImplementedError(
            f"{type(self)} does not implement abstract method #load_widget"
        )

    # =================================================================================== #
    # sibling methods                                                                     #
    # =================================================================================== #
    def next_sibling(self):
        return self._start_node.next_sibling()

    def prev_sibling(self):
        return self._start_node.prev_sibling()



from google.protobuf.descriptor_pb2 import (
    FileOptions,
)
from google.protobuf.internal.containers import (
    RepeatedCompositeFieldContainer,
    RepeatedScalarFieldContainer,
)
from google.protobuf.message import (
    Message,
)
from typing import (
    Iterable,
    List,
    Optional,
    Text,
    Tuple,
    cast,
)


class MethodOpt1(int):

    @classmethod
    def Name(cls, number: int) -> bytes: ...

    @classmethod
    def Value(cls, name: bytes) -> MethodOpt1: ...

    @classmethod
    def keys(cls) -> List[bytes]: ...

    @classmethod
    def values(cls) -> List[MethodOpt1]: ...

    @classmethod
    def items(cls) -> List[Tuple[bytes, MethodOpt1]]: ...


METHODOPT1_VAL1: MethodOpt1
METHODOPT1_VAL2: MethodOpt1


class AggregateEnum(int):

    @classmethod
    def Name(cls, number: int) -> bytes: ...

    @classmethod
    def Value(cls, name: bytes) -> AggregateEnum: ...

    @classmethod
    def keys(cls) -> List[bytes]: ...

    @classmethod
    def values(cls) -> List[AggregateEnum]: ...

    @classmethod
    def items(cls) -> List[Tuple[bytes, AggregateEnum]]: ...


VALUE: AggregateEnum


class TestMessageWithCustomOptions(Message):

    class AnEnum(int):

        @classmethod
        def Name(cls, number: int) -> bytes: ...

        @classmethod
        def Value(cls, name: bytes) -> TestMessageWithCustomOptions.AnEnum: ...

        @classmethod
        def keys(cls) -> List[bytes]: ...

        @classmethod
        def values(cls) -> List[TestMessageWithCustomOptions.AnEnum]: ...

        @classmethod
        def items(cls) -> List[Tuple[bytes,
                                     TestMessageWithCustomOptions.AnEnum]]: ...
    ANENUM_VAL1: TestMessageWithCustomOptions.AnEnum
    ANENUM_VAL2: TestMessageWithCustomOptions.AnEnum
    field1: Text
    oneof_field: int

    def __init__(self,
                 field1: Optional[Text] = ...,
                 oneof_field: Optional[int] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> TestMessageWithCustomOptions: ...


class CustomOptionFooRequest(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionFooRequest: ...


class CustomOptionFooResponse(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionFooResponse: ...


class CustomOptionFooClientMessage(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionFooClientMessage: ...


class CustomOptionFooServerMessage(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionFooServerMessage: ...


class DummyMessageContainingEnum(Message):

    class TestEnumType(int):

        @classmethod
        def Name(cls, number: int) -> bytes: ...

        @classmethod
        def Value(cls, name: bytes) -> DummyMessageContainingEnum.TestEnumType: ...

        @classmethod
        def keys(cls) -> List[bytes]: ...

        @classmethod
        def values(cls) -> List[DummyMessageContainingEnum.TestEnumType]: ...

        @classmethod
        def items(cls) -> List[Tuple[bytes,
                                     DummyMessageContainingEnum.TestEnumType]]: ...
    TEST_OPTION_ENUM_TYPE1: DummyMessageContainingEnum.TestEnumType
    TEST_OPTION_ENUM_TYPE2: DummyMessageContainingEnum.TestEnumType

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> DummyMessageContainingEnum: ...


class DummyMessageInvalidAsOptionType(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> DummyMessageInvalidAsOptionType: ...


class CustomOptionMinIntegerValues(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionMinIntegerValues: ...


class CustomOptionMaxIntegerValues(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionMaxIntegerValues: ...


class CustomOptionOtherValues(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> CustomOptionOtherValues: ...


class SettingRealsFromPositiveInts(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> SettingRealsFromPositiveInts: ...


class SettingRealsFromNegativeInts(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> SettingRealsFromNegativeInts: ...


class ComplexOptionType1(Message):
    foo: int
    foo2: int
    foo3: int
    foo4: RepeatedScalarFieldContainer[int]

    def __init__(self,
                 foo: Optional[int] = ...,
                 foo2: Optional[int] = ...,
                 foo3: Optional[int] = ...,
                 foo4: Optional[Iterable[int]] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> ComplexOptionType1: ...


class ComplexOptionType2(Message):

    class ComplexOptionType4(Message):
        waldo: int

        def __init__(self,
                     waldo: Optional[int] = ...,
                     ) -> None: ...

        @classmethod
        def FromString(
            cls, s: bytes) -> ComplexOptionType2.ComplexOptionType4: ...
    baz: int

    @property
    def bar(self) -> ComplexOptionType1: ...

    @property
    def fred(self) -> ComplexOptionType2.ComplexOptionType4: ...

    @property
    def barney(
        self) -> RepeatedCompositeFieldContainer[ComplexOptionType2.ComplexOptionType4]: ...

    def __init__(self,
                 bar: Optional[ComplexOptionType1] = ...,
                 baz: Optional[int] = ...,
                 fred: Optional[ComplexOptionType2.ComplexOptionType4] = ...,
                 barney: Optional[Iterable[ComplexOptionType2.ComplexOptionType4]] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> ComplexOptionType2: ...


class ComplexOptionType3(Message):

    class ComplexOptionType5(Message):
        plugh: int

        def __init__(self,
                     plugh: Optional[int] = ...,
                     ) -> None: ...

        @classmethod
        def FromString(
            cls, s: bytes) -> ComplexOptionType3.ComplexOptionType5: ...
    qux: int

    @property
    def complexoptiontype5(self) -> ComplexOptionType3.ComplexOptionType5: ...

    def __init__(self,
                 qux: Optional[int] = ...,
                 complexoptiontype5: Optional[ComplexOptionType3.ComplexOptionType5] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> ComplexOptionType3: ...


class ComplexOpt6(Message):
    xyzzy: int

    def __init__(self,
                 xyzzy: Optional[int] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> ComplexOpt6: ...


class VariousComplexOptions(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> VariousComplexOptions: ...


class AggregateMessageSet(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> AggregateMessageSet: ...


class AggregateMessageSetElement(Message):
    s: Text

    def __init__(self,
                 s: Optional[Text] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> AggregateMessageSetElement: ...


class Aggregate(Message):
    i: int
    s: Text

    @property
    def sub(self) -> Aggregate: ...

    @property
    def file(self) -> FileOptions: ...

    @property
    def mset(self) -> AggregateMessageSet: ...

    def __init__(self,
                 i: Optional[int] = ...,
                 s: Optional[Text] = ...,
                 sub: Optional[Aggregate] = ...,
                 file: Optional[FileOptions] = ...,
                 mset: Optional[AggregateMessageSet] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> Aggregate: ...


class AggregateMessage(Message):
    fieldname: int

    def __init__(self,
                 fieldname: Optional[int] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> AggregateMessage: ...


class NestedOptionType(Message):

    class NestedEnum(int):

        @classmethod
        def Name(cls, number: int) -> bytes: ...

        @classmethod
        def Value(cls, name: bytes) -> NestedOptionType.NestedEnum: ...

        @classmethod
        def keys(cls) -> List[bytes]: ...

        @classmethod
        def values(cls) -> List[NestedOptionType.NestedEnum]: ...

        @classmethod
        def items(cls) -> List[Tuple[bytes, NestedOptionType.NestedEnum]]: ...
    NESTED_ENUM_VALUE: NestedOptionType.NestedEnum

    class NestedMessage(Message):
        nested_field: int

        def __init__(self,
                     nested_field: Optional[int] = ...,
                     ) -> None: ...

        @classmethod
        def FromString(cls, s: bytes) -> NestedOptionType.NestedMessage: ...

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> NestedOptionType: ...


class OldOptionType(Message):

    class TestEnum(int):

        @classmethod
        def Name(cls, number: int) -> bytes: ...

        @classmethod
        def Value(cls, name: bytes) -> OldOptionType.TestEnum: ...

        @classmethod
        def keys(cls) -> List[bytes]: ...

        @classmethod
        def values(cls) -> List[OldOptionType.TestEnum]: ...

        @classmethod
        def items(cls) -> List[Tuple[bytes, OldOptionType.TestEnum]]: ...
    OLD_VALUE: OldOptionType.TestEnum
    value: OldOptionType.TestEnum

    def __init__(self,
                 value: OldOptionType.TestEnum,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> OldOptionType: ...


class NewOptionType(Message):

    class TestEnum(int):

        @classmethod
        def Name(cls, number: int) -> bytes: ...

        @classmethod
        def Value(cls, name: bytes) -> NewOptionType.TestEnum: ...

        @classmethod
        def keys(cls) -> List[bytes]: ...

        @classmethod
        def values(cls) -> List[NewOptionType.TestEnum]: ...

        @classmethod
        def items(cls) -> List[Tuple[bytes, NewOptionType.TestEnum]]: ...
    OLD_VALUE: NewOptionType.TestEnum
    NEW_VALUE: NewOptionType.TestEnum
    value: NewOptionType.TestEnum

    def __init__(self,
                 value: NewOptionType.TestEnum,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> NewOptionType: ...


class TestMessageWithRequiredEnumOption(Message):

    def __init__(self,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> TestMessageWithRequiredEnumOption: ...



from collections import defaultdict
from enum import Enum

import urwid
from loguru import logger
from overrides import overrides

from .common import PopUpLauncher
from .components import JSONBrowser, QueryBar, HelpBar, AutoCompletePopUp


class ViewFrame(PopUpLauncher):
    """
    A wrapper of the frame as the main UI of `pyfx`.
    """

    def __init__(self, client, view_manager, keymapper):
        self._manager = view_manager
        self._keymapper = keymapper
        self._handlers = defaultdict(list)

        self._json_browser = JSONBrowser(self, keymapper.json_browser)
        self._query_bar = QueryBar(self, client, keymapper.query_bar)
        self._help_bar = HelpBar(self)

        super().__init__(urwid.Frame(self._json_browser, footer=self._help_bar))
        # json browser
        self.register("open_query_bar", self.focus_on_query)
        # autocomplete
        self.register("close_pop_up", self.close_pop_up)
        # query bar
        self.register("open_pop_up", self.open_pop_up)
        self.register("focus_on_view", self.focus_on_view)
        self.register("get_component_size", self.size)

    def notify(self, signal, source, *args, **kwargs):
        """
        Broadcast signals to all the listeners registered for the signal and
        collect result.
        """
        try:
            results = []
            for callback in self._handlers[signal]:
                results.append(callback(*args, **kwargs))
            return results
        except KeyError:
            logger.opt(exception=True) \
                .warning(f"Key ({source}, {signal}) is not defined")

    def register(self, signal, callback):
        """
        Handlers defines a set of signals this view frame will broadcast to,
        it is the callback's responsibility to filter out unused signal based on
        passed source.
        """
        self._handlers[signal].append(callback)

    def size(self):
        return self._manager.size()

    def set_data(self, data):
        self._json_browser.set_top_node(data)

    def focus_on_view(self):
        if self.original_widget.body is not self._json_browser:
            self._change_widget(self._json_browser, FocusArea.BODY)
        self._change_focus(FocusArea.BODY)

    def focus_on_query(self):
        if self.original_widget.footer is not self._query_bar:
            self._query_bar.setup()
            self._change_widget(self._query_bar, FocusArea.FOOTER)
        self._change_focus(FocusArea.FOOTER)

    def _change_widget(self, widget, area):
        if area == FocusArea.BODY:
            self.original_widget.body = widget
        elif area == FocusArea.FOOTER:
            self.original_widget.footer = widget
        else:
            # swallow this error but log warnings
            logger.warning(
                "Unknown area {} for switching widgets.",
                area.value
            )

    def _change_focus(self, area):
        self.original_widget.focus_position = area.value

    # This is a workaround we used to be able to popup autocomplete window in
    # query bar
    # The implementation of PopUpLauncher only support pop up within the
    # launcher's canvas, i.e., autocomplete-edit's popup launcher should be
    # implemented in the container widget of the edit widget
    @overrides
    def create_pop_up(self, *args, **kwargs):
        return AutoCompletePopUp(
            self, self._keymapper.autocomplete_popup, *args, **kwargs
        )

    @overrides
    def get_pop_up_parameters(self, size, *args, **kwargs):
        cur_col, _ = self.original_widget.get_cursor_coords(size)
        popup_max_col, popup_max_row = self.pop_up_widget.pack(size)
        max_col, max_row = size
        footer_rows = self.original_widget.footer.rows((max_col,))
        return {
            'left': cur_col,
            'top': max_row - popup_max_row - footer_rows,
            'overlay_width': popup_max_col,
            'overlay_height': popup_max_row
        }


class FocusArea(Enum):
    """
    Enum for focus area in Main Window
    """
    BODY = "body"
    FOOTER = "footer"



from __future__ import unicode_literals

from rp.prompt_toolkit.completion import Completer, Completion
from rp.prompt_toolkit.document import Document
from rp.prompt_toolkit.contrib.completers import PathCompleter
from rp.prompt_toolkit.contrib.regular_languages.compiler import compile as compile_grammar
from rp.prompt_toolkit.contrib.regular_languages.completion import GrammarCompleter

from rp.rp_ptpython.utils import get_jedi_script_from_document
# from rp import printed,fansi_print
import re

__all__ = (
    'PythonCompleter',
)
non_completable_keywords={'class','finally','is','return','continue','for','lambda','try','def','from','nonlocal','while','and','del','global','not','with','as','elif','if','or','yield','assert','else','import','pass','break','except','in','raise'}

import sys
_all_module_names=set(sys.builtin_module_names)
def get_all_importable_module_names():
    #Returns a set of all names that you can use 'import <name>' on
    import pkgutil
    if _all_module_names:
        def update_thread():
            #Every time we call this fuction, make a new thread that will update the _all_module_names for us (without blocking)
            #That way we can respond to new packages being added, or changes in directory etc without having to block completions for them
            for _,name,_ in pkgutil.iter_modules():
                _all_module_names.add(name)
        from threading import Thread
        Thread(target=update_thread).start()
        return _all_module_names
    for _,name,_ in pkgutil.iter_modules():
        _all_module_names.add(name)
    return _all_module_names

from threading import Thread
Thread(target=get_all_importable_module_names).start()
Thread(target=lambda: __import__('rp').r._get_cached_system_commands).start()


MAX_SORTING_PRIORITY=999999999
sorting_priorities={}
    

_get_apt_completions_cache=None
def get_apt_completions():
    #Gets all installable apt packages. Perhaps migrate this to rp?
    global _get_apt_completions_cache
    if _get_apt_completions_cache is not None:
        return _get_apt_completions_cache.copy()
    else:
        import subprocess
        # Update package lists with 'sudo apt update'
        # subprocess.run(['sudo', 'apt', 'update'], stdout=subprocess.PIPE)
        
        try:
            # Get the list of all available packages
            result = subprocess.run(['apt-cache', 'dumpavail'], stdout=subprocess.PIPE, text=True)
            
            # Decode the byte string into a string, split by newline
            package_list = result.stdout.split('\n')
            
            # Extract package names
            packages = set()
            for line in package_list:
                if line.startswith('Package: '):
                    package_name = line.split(' ')[1]
                    packages.add(package_name)
            
            _get_apt_completions_cache= packages
        except Exception:
            _get_apt_completions_cache=[]

        return _get_apt_completions_cache

SHOWN_ERROR=False
class PythonCompleter(Completer):
    """
    Completer for Python code.
    """
    def __init__(self, get_globals, get_locals):
        super(PythonCompleter, self).__init__()

        self.get_globals = get_globals
        self.get_locals = get_locals

        self._path_completer_cache = None
        self._path_completer_grammar_cache = None

    @property
    def _path_completer(self):
        if self._path_completer_cache is None:
            self._path_completer_cache = GrammarCompleter(
                self._path_completer_grammar, {
                    'var1': PathCompleter(expanduser=True),
                    'var2': PathCompleter(expanduser=True),
                })
        return self._path_completer_cache

    @property
    def _path_completer_grammar(self):
        """
        Return the grammar for matching paths inside strings inside Python
        code.
        """
        # We make this lazy, because it delays startup time a little bit.
        # This way, the grammar is build during the first completion.
        if self._path_completer_grammar_cache is None:
            self._path_completer_grammar_cache = self._create_path_completer_grammar()
        return self._path_completer_grammar_cache

    def _create_path_completer_grammar(self):
        def unwrapper(text):
            return re.sub(r'\\(.)', r'\1', text)

        def single_quoted_wrapper(text):
            return text.replace('\\', '\\\\').replace("'", "\\'")

        def double_quoted_wrapper(text):
            return text.replace('\\', '\\\\').replace('"', '\\"')

        grammar = r"""
                # Text before the current string.
                (
                    [^'"#]                                  |  # Not quoted characters.
                    '''  ([^'\\]|'(?!')|''(?!')|\\.])*  ''' |  # Inside single quoted triple strings
                    "" " ([^"\\]|"(?!")|""(?!^)|\\.])* "" " |  # Inside double quoted triple strings

                    \#[^\n]*(\n|$)           |  # Comment.
                    "(?!"") ([^"\\]|\\.)*"   |  # Inside double quoted strings.
                    '(?!'') ([^'\\]|\\.)*'      # Inside single quoted strings.

                        # Warning: The negative lookahead in the above two
                        #          statements is important. If we drop that,
                        #          then the regex will try to interpret every
                        #          triple quoted string also as a single quoted
                        #          string, making this exponentially expensive to
                        #          execute!
                )*
                # The current string that we're completing.
                (
                    ' (?P<var1>([^\n'\\]|\\.)*) |  # Inside a single quoted string.
                    " (?P<var2>([^\n"\\]|\\.)*)    # Inside a double quoted string.
                )
        """

        return compile_grammar(
            grammar,
            escape_funcs={
                'var1': single_quoted_wrapper,
                'var2': double_quoted_wrapper,
            },
            unescape_funcs={
                'var1': unwrapper,
                'var2': unwrapper,
            })

    def _complete_path_while_typing(self, document):
        char_before_cursor = document.char_before_cursor
        return document.text and (
            char_before_cursor.isalnum() or char_before_cursor in '/.~')

    def _complete_python_while_typing(self, document):
        char_before_cursor = document.char_before_cursor
        return document.text and (
            char_before_cursor.isalnum() or char_before_cursor in '_.')

    def _get_completions(self, document, complete_event,force=False):# When force is true it acts like the original ptpython autocomplete
      sorting_priorities.clear()
      import warnings
      with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        force=True
        global old_origin,candidates
        origin=document.get_word_before_cursor()

        def yield_from_candidates(candidates:list,priority=None,displays=None):
            candidates=list(candidates)
            if priority is None:
                priority=[MAX_SORTING_PRIORITY]*len(candidates)
            if displays is None:
                displays=candidates
            for c,p,d in zip(candidates,priority,displays):
                #TODO: RIGHT NOW displays IS IGNORED BY get_completions WHICH USES THIS FUNCTION. ANNOYING!
                completion=Completion(text=c,start_position=-len(origin),display=d);
                sorting_priorities[c]=p
                # print(c,p)
                yield completion
        import rp.r_iterm_comm as ric
        current_line=document.current_line
        before_line=document.current_line_before_cursor
        after_line=document.current_line_after_cursor
        after=document.text_after_cursor
        before=document.text_before_cursor


        import re
        from rp import ring_terminal_bell
        # ring_terminal_bell()
        if re.fullmatch(r' *(from|import) +\w*',before_line):
            #Speed up 'import ___|' completions. Note that this is going to skip modules that have been pip_installed during runtime, but this is a small price to pay for this kind of crazy fast autocompletion speed.
            #This kinda completion is particularly useful because the spacebar triggers the autocpmpletion of module names like tab does
            yield from yield_from_candidates(get_all_importable_module_names())
            return
        # if re.fullmatch(r'.*[^\w\.]',before_line):
        #     yield from yield_from_candidates(list(ric.globa))
        #     # from rp import play_tone
        #     # play_tone(880,seconds=.03)  
        #     return
        if not '\n' in document.text and re.fullmatch(r' *((if|with|while|assert|yield|yield from|elif) )? *\w*',before_line):
            # ring_terminal_bell()
            yield from yield_from_candidates(list(ric.globa))
            return
        blr=before_line[::-1]#blr=before_line reversed
        if '.' in blr:#This is a hack to make completions to namespaces like np.* not need to use jedi (which takes about 1 second to make a script which is super slow). This is a hack but it makes completions that would otherwise have to use jedi over 20x faster.
            found=re.match(r'(\.\w*[a-zA-Z_])+',blr[blr.find('.'):])
            if not found:
                pass#Something went wrong whatever
            elif found.string:
                found=found.string
                assert found,'wat? found='+repr(found.match)
                prefix=found[1:][::-1]
                try:
                    def keys(root):
                        out=set()
                        try:out.update(dir(root))
                        except:pass
                        try:out.update(root.__dict__)
                        except:pass
                        return sorted(out)
                    def evall():
                        parts=prefix.split('.')
                        cursor=ric.globa[parts[0]]
                        for part in parts[1:]:
                            cursor=getattr(cursor,part)
                        return cursor
                    yield from yield_from_candidates(keys(evall()))
                    # yield from yield_from_candidates(keys(eval(prefix,ric.globa)))
                    # from rp import play_tone
                    # play_tone(seconds=.03)
                    # print("HI"+repr(prefix)+"BYE")
                    return
                except Exception as E:
                    # from rp import print_stack_trace
                    # print_stack_trace(E)
                    #This type of error is perfectly normal and is to be expected.
                    #  ⮤ adsfpor.print_stack_trace: ERROR: Traceback (most recent call last):
                    #   File "/Library/Frameworks/Python.framework/Versions/3.5/lib/python3.5/site-packages/rp/rp_ptpython/completer.py", line 136, in _get_completions                                                                                     
                    #     yield from yield_from_candidates(keys(eval(prefix,ric.globa)))                                  
                    #   File "<string>", line 1, in <module>                                                              
                    # NameError: name 'adsfpo' is not defined  
                    pass#Oh well whatevs we'll do it with jedi, it will just be a bit slower

        import re
        current_line=document.current_line
        before_line=document.current_line_before_cursor
        after_line=document.current_line_after_cursor
        after=document.text_after_cursor
        before=document.text_before_cursor
        in_global_or_nonlocal_declaration=(re.fullmatch(r'\s*(del|global|nonlocal)\s+[\w\s\,]*',before_line))
        # if in_global_or_nonlocal_declaration:
        #     print("GOOGOO")
            #Make @memoized complete faster:
            #If we set the mode to 'fast' instead of 'good':
            #after 'if |' or 'while |' etc, you might want to use fast autocompletions instead of good autocompletions, because I can't think of a situation where it would miss any completions (though it might rip some out of strings and comments). This gives a speed boost.
        # if re.fullmatch(r'\s*def\s+\w+\s*\((\s*\w+\s*(\=\s*[^\s].*)?\s*\,\s*)*\w*',before_line) and after_line.strip()=='):':
        #     return#Don't autocomplete function paramater names

        from rp import starts_with_any

        def pathmod(entry):
            if isinstance(entry,str):
                if path_before_cursor is not None:
                    #For completing CD and TAKE etc
                    entry=entry[len(path_before_cursor):]
                elif entry.startswith('./'): 
                    #For completing in strings primarily
                    #This is just bloat
                    entry =entry[len('./'):]
                return entry
            out = entry.path
            if entry.is_dir():
                out+='/'
            assert isinstance(out,str)
            return pathmod(out)

        def pathsmod(entries,priority=lambda x:MAX_SORTING_PRIORITY):
            mods=[]
            pris=[]
            for x in entries:
                mods.append(pathmod(x))
                pris.append(priority(x))
            return mods,pris

        path_before_cursor=None
        def get_path_before_cursor():
            nonlocal path_before_cursor #Cache this

            #Assume we're like 'TAKE /path'  or  'CAT ./path' etc - we have a space after the first word
            assert ' ' in before_line
            first_word=before_line.split()[0]
            line=before_line
            line=line[len(first_word):]
            line=line.lstrip()
            if not line:
                path= '.'
            elif line.endswith('/'):
                path= line
            elif '/' in line:
                path= '/'.join(line.split('/')[:-1])
            else:
                path= line

            import rp
            if rp.path_exists(path):
                path_before_cursor = path
            else:
                #Don't crash, hack here. No reason to return '.', really, other than that. Cant pass bad path to scandir.
                path_before_cursor = '.'

            return path_before_cursor


        def scandir_path_before_cursor():
            try:
                return os.scandir(get_path_before_cursor())
            except Exception:
                #Please ignore for  now
                pass
                
        if starts_with_any(before_line,'CDH ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            import rp

            updirs = rp.r._get_cdh_back_names()
                
            # yield from yield_from_candidates(updirs)
            yield from yield_from_candidates(updirs,displays=updirs,priority=list(range(len(updirs))))
            return 
        if starts_with_any(before_line,'CD ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import rp
            import os
            from rp import is_a_directory
            yield from yield_from_candidates(*pathsmod(x for x in scandir_path_before_cursor() if x.is_dir()))
            return 
        if starts_with_any(before_line,'TAKE ','MKDIR ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            from rp import is_a_directory
            yield from yield_from_candidates(*pathsmod(scandir_path_before_cursor(), priority= lambda x:not x.is_dir()))
            return 
        if starts_with_any(before_line,'PYM ','APYM ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            def get_module_names():
                import pkgutil
                for _, module_name, _ in pkgutil.iter_modules():
                    yield module_name
            yield from yield_from_candidates(get_module_names())
            return 
        if starts_with_any(before_line,'CDU ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            import rp

            pwd=rp.get_current_directory()
            cwd=pwd
            displays=[]
            updirs=[]
            priorities=[]
            for _ in range(len(rp.path_split(pwd))):
                cwd=rp.get_parent_directory(cwd)
                if rp.get_folder_name(cwd):
                    priorities.append(-len(cwd))
                    upname=rp.get_folder_name(cwd)
                    displays.append(upname)
                    updirs.append(upname)
                    # updirs.append('/'+cwd)

                
            yield from yield_from_candidates(updirs,displays=displays,priority=priorities)
            return 
        if starts_with_any(before_line,'RN ','RM ','VIM ','OPEN ','FD ','WANS ','MV ','PIP ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            yield from yield_from_candidates(*pathsmod(scandir_path_before_cursor()))
            return 
        if re.fullmatch(r'(CAT |NCAT |CCAT |OPEN |ACAT |TAB |RUN |PY |APY |PYM |APYM |LSS |LSR |FCOPY ).*',before) and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            from rp import is_a_file
            import rp
            from rp import printed

            #THIS WORKS but it interferes with pterm getting the completions for completing LSS without tab
            # after_cmd = '' if len(before.split())<2 else before.split()[1] #If we use a . in the command prioritize hidden path autocompletions 

            yield from yield_from_candidates(
                *pathsmod(
                    scandir_path_before_cursor(),
                    priority=lambda x: (
                        x.is_dir() #  + (2 * x.path[len("./") :].startswith(".")) * (not after_cmd.startswith('.'))  #Also commented out because it interferes with enter-completions from pterm LSS
                    ),
                )
            )  # Put directories last but do include them
            return 
        if re.fullmatch(r'.*\`\w*',before_line)\
            or (before_line.startswith('RUN ') and not ('\n' in before) and not after):#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            yield from yield_from_candidates([x for x in os.listdir() if any(x.endswith(e) for e in '.py .rpy'.split())])
            return 
        if (before_line.startswith('!') or before_line.startswith('ARG ')) and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
            import os
            bls=before_line.split()
            if before_line.startswith('!sudo apt install'):
                yield from yield_from_candidates(get_apt_completions())
                return
            if before_line.startswith("!") and not (
                    before_line[1:].startswith("/") or before_line[1:].startswith(".")
                ) and (
                    len(bls) == 1
                    and (before_line.strip() == before_line or bls[0]=='!sudo')
                    or len(bls) == 2
                    and bls[0] == "!sudo"
                    and (not bls[1].startswith(".") or bls[1].startswith("/"))
                    and before_line==before_line.strip()
                ):
                    import rp
                    yield from yield_from_candidates(rp.r._get_cached_system_commands())
            else:
                yield from yield_from_candidates([x for x in os.listdir()])
            return 

        import re

        # if re.fullmatch(r'(.*\W)?',before_line)


        if  in_global_or_nonlocal_declaration     or\
            (not after and re.fullmatch(r'\@\w*',before_line)) or\
            (re.fullmatch(r'\s*(if|while|(for\s+(\w\s\,)+\s+in)|assert|with|return|yield|(yield from)|elif|)\s+\w*',before_line)) or \
            (ric.completion_style and ric.completion_style[0]=='fast'): #fast but bad completion mode
            tokens=re.findall(r'[a-zA-Z_]\w*',document.text)#all possible tokens in the current buffer
            yield from yield_from_candidates((set(tokens)|set(ric.globa))-non_completable_keywords)
            return


        ##################### JEDI SECTION. THIS IS A SLOW AREA  ##############

        # ric.current_candidates.clear()
        # if not origin:
        #     return
        if old_origin not in origin or force:# Make things faster when we're just adding on to a previous autocomplete thing by reusing our previous autocomplete data

            """
            Get Python completions.
            """
            # Do Path completions
            if complete_event.completion_requested or self._complete_path_while_typing(document):
                for c in self._path_completer.get_completions(document, complete_event):
                    # ric.current_candidates.append(c)
                    yield c

            # If we are inside a string, Don't do Jedi completion.
            import rp
            if self._path_completer_grammar.match(document.text_before_cursor):
                rp.r_iterm_comm.writing_in_string=True
                return
            rp.r_iterm_comm.writing_in_string=False

            # Do Jedi Python completions.
            if complete_event.completion_requested or self._complete_python_while_typing(document):
                
                cursor_position=document.cursor_position

                if not '/' in origin:
                    #If we're completing a path, don't modify this.
                    #Otherwise, please grab as many completions as we can so we can efficiently backtrack through them
                    cursor_position-=(origin[::-1].find(".") if "." in origin else len(origin))   #NOT SURE WHAT THIS WAS FOR (april 19 2024) --- Disabling it seems to do nothing?? OH wait...I think I know what this is for...its to make sure we get ALL the completions so our code can handle getting subsets faster. Re-enabling.

                doc = Document(
                    document.text,
                    cursor_position,
                    document.selection,
                )

                script = get_jedi_script_from_document(doc, self.get_locals(), self.get_globals())
                # print(script)
                pos=document.cursor_position-len(origin)+1
                if script and not isinstance(script,Exception):
                    try:completions = script.completions()
                    except Exception as e:
                        import rp
                        # rp.print_verbose_stack_trace(e)
                        pass#The original ptpython has a whole string of specific error types, and it looks like visual clutter. I got rid of it to see my code more easily. I dont think im missing much.
                    else:
                        # fff=1
                        for c in completions:
                            # if fff:
                                # from rp import ring_terminal_bell
                                # ring_terminal_bell()
                                # fff=0
                            thingy=Completion(c.name_with_symbols, len(c.complete) - len(c.name_with_symbols),display=c.name_with_symbols)
                            yield thingy
                        # else:
                            # from rp import ring_terminal_bell
                        return
                        candidates={c.name_with_symbols for c in completions}
                else:
                    def update_parso_grammars():
                        #Get all the latest grammar files but don't update parso as a whole
                        #This will make rp's autocomplete features happy
                        import rp
                        old_dir=rp.get_current_directory()
                        rp.set_current_directory(rp.get_parent_directory(rp.get_module_path_from_name('rp.libs.parso.python')))
                        ans=['https://raw.githubusercontent.com/davidhalter/parso/master/parso/python/grammar%s.txt'%v for v in '38 39 310 311 312 313 314 315 316 317 318 319 320'.split()]
                        for x in ans:
                            import os
                            os.system('wget -N '+x)
                        print('\n\nSaved some new files to %s'%rp.get_current_directory())
                        rp.set_current_directory(old_dir)
                        print('\n\n\n'+'Done! You may continue using rp without this error (hopefully).')
                        
                    print("ATTENTION: Please update the parso definitions! This message came from %s, and script="%__file__+repr(script))
                    if rp.input_yes_no("rp's autocomplete functionality is broken because it doesn't recognize this version of Python's grammar. Is it ok if we try and fix this by downloading some up-to-date files from https://github.com/davidhalter/parso?"):
                        update_parso_grammars()
                    else:
                        print("OK...I'm assuming you said NO because you tried this and it didn't fix the problem, so here's a big stack trace to help you debug it. If this is annoying you could comment it out in %s, but note that autocomplete won't work fully if you do that. Here's the stack trace:"%__file__)
                        if True: #Set this to False to disable the stack trace during errors
                            import rp
                            if isinstance(script,Exception):
                                rp.print_verbose_stack_trace(script)
                            print('doc='+str(doc),'\n\nscript='+repr(script),len(self.get_locals()))
                candidates=set(candidates)
                # if not force:
                #     candidates|=set(self.get_globals())|set(self.get_locals())# The namespace we search for completions in
        candidates =ryan_completion_matches(origin,tuple(candidates))    # Now its a list, and we've ordered our search.
        for c in candidates:
            # ric.current_candidates.append(c);
            yield Completion(text=c,start_position=-len(origin),display=c);
        else:
            from rp import is_namespaceable
            if not force and (not is_namespaceable(origin) or not origin):
                for x in self.get_completions(Document(document.text,document.cursor_position-len(origin)+(1 and len(origin)),),complete_event,force=True):
                    # ric.current_candidates.append(x);
                    yield x;# For when we want autocompletins for 'np.' or 'rinsp.' or 'rp.' etc, as opposed to 'pr' to print or 'lis' to list
        # if not force and not ric.current_candidates or origin not in old_origin and origin!=old_origin:old_origin=origin;return self.get_completions(document, complete_event,force=True)

        old_origin=origin
    def get_completions(self, document, complete_event,force=False):
     import rp
     try:
      import warnings
      with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        global completion_cache_pre_origin_doc
        from rp import tic,toc,ptoctic,ptoc 
        # tic()
        origin=document.get_word_before_cursor()
        import rp.r_iterm_comm as ric
        pre_origin_doc=document.text[:document.cursor_position-(origin[::-1].find('.') if '.' in origin else len(origin))]
        from rp import ring_terminal_bell
        # if pre_origin_doc.endswith('.'):
            # ring_terminal_bell()
        # print()
        # print(pre_origin_doc)
        # print()
        # document=buffer.document

        before_line=document.current_line_before_cursor
        current_line=document.current_line
        after_line=document.current_line_after_cursor
        after=document.text_after_cursor
        before=document.text_before_cursor

        if not after_line and set(before_line)==set('u'):#not after and not '\n' in before and re.fullmatch(before_line):
            #Complete uuuuuu --> CDU that dir   (lets us see where we're going in the completions! No need to count)
            import os
            import rp

            cwd=rp.get_current_directory()
            for _ in range(len(before_line)):
                cwd=rp.get_parent_directory(cwd)

            # updir='CDU '+rp.get_folder_name(cwd)
            updir='CD '+cwd

                
            yield Completion(text=updir, start_position=-10000, display='^'+rp.get_folder_name(cwd))

        if not before_line.strip() and not force:
           return

        ##This paragraph: CD path completion. This should probably be in _get_completions for efficieny with large dicts, but for now this works.
        #before=document.text_before_cursor
        #after=document.text_after_cursor
        #def cd_path_completion_yield_from_candidates(candidates:list):
        #    for c in candidates:
        #        yield Completion(text=c,start_position=-len(origin),display=c[1:]);
        #from rp import get_path_parent
        #def path_complete(path):
        #    path=path.strip()
        #    if path.endswith('/'):
        #        return os.listdir(path)
        #    else:
        #        return os.listdir(get_path_parent(path))
        #if before_line.startswith('CD ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
        #    import os
        #    from rp import is_a_directory
        #    paths=before_line[len('CD '):]
        #    paths=path_complete(paths)
        #    yield from cd_path_completion_yield_from_candidates([x for x in paths if is_a_directory(x)])
        #    return 

        ##This paragraph: Dictionary key completion. This should probably be in _get_completions for efficieny with large dicts, but for now this works.
        ##This completes keys in a dictionary iff the dict is a global variable, and it's the first thing you've typed in the buffer
        ##(it's pretty limited - and only does stuff if all the keys are python literals)
        ##For example, if d=={'Hello':'World','Goodbye':'Sam'} then we get completions like this:
        ##     >>> d[|]
        ##         'Hello'
        ##         'Goodbye'
        #def dict_completion_yield_from_candidates(candidates:list):
        #    for c in candidates:
        #        yield Completion(text=c,start_position=-len(origin),display=c[1:]);
        #after_line=document.current_line_after_cursor
        #def is_any_instance(x,*types):
        #    for t in types:
        #        if isinstance(x,t):
        #            return True
        #    return False
        #from collections import OrderedDict
        #if before_line.endswith('[') and after_line==']':
        #    if before_line[:-1] in ric.globa:
        #        if is_any_instance(ric.globa[before_line[:-1]],dict, OrderedDict):
        #           if all(is_any_instance(x,str,int,float,bool) for x in ric.globa[before_line[:-1]]):
        #            # assert False
        #            dict_completion_candidates=[]
        #            for x in ric.globa[before_line[:-1]]:
        #                dict_completion_candidates.append('['+repr(x))
        #            ric.current_candidates[:]=dict_completion_candidates
        #            yield from dict_completion_yield_from_candidates(candidates)
        #            return

        flag=False
        while True:
            if pre_origin_doc not in completion_cache_pre_origin_doc or flag:
                ric.current_candidates=[]
                from rp import tic,ptoctic
                # tic()
                for c in self._get_completions(document, complete_event,force=False):
                    # print(c.text)
                    # print(c.text)
                    ric.current_candidates.append(str(c.text))
                # ptoctic()
                completion_cache_pre_origin_doc[pre_origin_doc]=tuple(ric.current_candidates)
                break
            else:
                ric.current_candidates=list(completion_cache_pre_origin_doc[pre_origin_doc])
                if not ric.current_candidates:
                    from rp import ring_terminal_bell
                    flag=True
                    # ring_terminal_bell()
                    continue

                else:
                    break
        # print("TIME1:",end='')
        # ptoc()
        # print()
        # print(ric.current_candidates)
        post_period_origin=origin if '.' not in origin else origin[origin.find('.')+1:]
        ric.current_candidates=ryan_completion_matches(post_period_origin,tuple(ric.current_candidates))
        # if '.' in origin:
            # ric.current_candidates==ric.current_candidates[::-1]

        user_created_var_names = set(ric.rp_pt_user_created_var_names)

        
        def sorting_key(x):
            # Assigns a sort key based on prefix rules:
            # 1. Items starting with '.' come last.
            # 2. Items starting with '_' come next.
            # 3. Other items come first.
            orig = x
            if x.startswith('.'):
                if origin.startswith('.'):
                    #If explicitly looking for hidden file paths, put it first
                    output = -1
                else:
                    #If not explicitly looking for that, put it last
                    output = 1
            elif x.startswith('_'):
                if origin.startswith('_'):
                    #If explicitly looking for privates, put it first
                    output = -1
                else:
                    #If not explicitly looking for that, put it last
                    output = 1
            else:
                output = 0  # Lowest sort value for other items

            #Put user-made vars first - that means not putting rp things first
            user_created_priority = 0 if x in user_created_var_names else 1

            output = (output, user_created_priority)
            output = (output, sorting_priorities.get(orig,MAX_SORTING_PRIORITY))

            return output

        out_completions=[]

        # print(sorting_priorities)

        # Single loop processing all candidates, sorted by the custom key
        for x in sorted(ric.current_candidates, key=sorting_key):
            if x.endswith("'") or x.endswith('"'): x=x[:-1] #Autocompleted paths in strings are a bit buggy right now

            #When completing 'System/|' origin is just '/'

            def unhidden_path(x):
                o=x
                x=x.strip()
                while x[0] in '/':
                    x=x[1:]
                if x.startswith('.'):
                    return x[1:]
                return o

            #TODO: THis was edited on MACX to give path completion for CD etc on April 19 2024. It's very sloppy and full of bugs and really ugly, done in los angeles. TODO: Undo this commit and make it prettier! 
            if x.startswith('/') and (not '/' in before_line or before_line.endswith('/')):
                x=x[1:]
                
            text=x
            display=x

            if x.strip().endswith('/'):
                #Its annoying when folders are completed with /'s because then when we type tab it goes away
                #But if we type / we get folder// instead.
                #Let the user type the /
                text=text.strip()[:-1]
                if '/' in origin:
                    #Make sure dividers are right
                    text = "/" + text

            if '.' in x and before_line.endswith('.'):
                #Don't complete .g to ..git   or   'rp/.' to 'rp/..git' make it 'rp/.git' instead
                text=unhidden_path(text)

            # display=origin #For debugging


            out_completions.append(Completion(text=text, start_position=-len(post_period_origin), display=display))

        yield from out_completions
     except Exception as e:
         global SHOWN_ERROR
         if not SHOWN_ERROR:
           print()
           rp.fansi_print("  >> RP AUTOCOMPLETION ERROR START : SAVED EXCEPTION TO rp.rp_ptpython.completer.SHOWN_ERROR <<  ",'red red bold on black black')
           rp.print_stack_trace()
           rp.fansi_print("  >> WILL ONLY SHOW COMPLETION ERRORS ONCE PER SESSION TO AVOID SPAM <<  \n  >> SET rp.rp_ptpython.SHOWN_ERROR=True IN RPRC TO DISABLE THESE COMPLETION ERROR REPORTS <<  ",'red red bold on black black')
           SHOWN_ERROR=e

        #OLD VERSION BEFORE SORTING_KEY
        # for x in ric.current_candidates:
        #     if not x.startswith('_') and not x.startswith('.'):#Make sure things starting with '_' come last...
        #         yield Completion(text=x,start_position=-len(post_period_origin),display=x)
        # for x in ric.current_candidates:
        #     if     x.startswith('_'):
        #         yield Completion(text=x,start_position=-len(post_period_origin),display=x)
        # for x in ric.current_candidates:
        #     if     x.startswith('.'): #Private files, such as .backup_.py should be displayed last
        #         yield Completion(text=x,start_position=-len(post_period_origin),display=x)

        # print("TIME2:",end='')
        # ptoc()
        # print()

        # return ric.current_candidates

completion_cache_pre_origin_doc={}#Using a bit of dynamic programming to speed up autocompletion by A FACTOR OF 7!! (I measured it with the commented tics/tocs)
#region Time for some super-speedy optimization tricks...
old_origin=''
candidates=set()

from functools import lru_cache

_ryan_completion_matches_cache={}#Note: over time this might cause a memory leakage....idk if that will be a problem....
def ryan_completion_matches(origin:str,candidates:list):
    import rp
    #This function accomplishes one of the few **beautiful** things in rp.pseudo_terminal: my string matching algorithm, specifically made for python namespace elements
    # Better autocompletions
    # candidates origin is string
    # candidates is list of strings
    # def is_match(origin:str,candidate:str):
    #     origin   =list(origin   )# So we can pop chars
    #     candidate=list(candidate)
    #     while origin and candidate:
    #         if origin[0]==candidate.pop(0):
    #             origin.pop(0)
    #     return not origin
    h=hash(origin),hash(candidates)
    if h in _ryan_completion_matches_cache:
        # print(end="HAHA");rp.ptoc()

        return _ryan_completion_matches_cache[h]

    candidates=list(candidates)

    def match(origin:str,candidate:str)->int:
        # from rp import regex_match
        # if not candidate.strip():
            # return None
        # try:
        # import regex
        # if origin and not :#Optimization to quickly weed out most candidates
            # return None
        # except:
            # print("ERROR AT origin",candidate)

        temp=candidate
        origin   =list(origin   )# So we can pop chars
        candidate=list(candidate)
        out=.001
        count=0# higher count means worse match
        def match_char(a:str,b:str):
            return a.upper()==b.upper()
        while origin and candidate:
            pop=candidate.pop(0)
            if match_char(origin[0],pop):
                org=origin.pop(0)
                if org!=pop:
                    count+=.0001# small penalty for wrong char case. still keep it but dont give it ordering priority
                out+=count
                count=0
            elif pop in '_':
                count=0
                out+=.1# penalty because 'ab' is closer to 'a_bz' than 'az_b'
            else:
                count+=1
        import rp.r_iterm_comm
        # rp.tic()
        try:
            temp=(''.join(rp.r_iterm_comm.successful_commands)).count(temp)
            if temp:
                out/=temp# Doesn't have to be division; this is arbitrary. Basically, if we used this before successfully give it more weight and likelyness we want to use it again
        except AttributeError:pass# AttributeError: module 'rp.r_iterm_comm' has no attribute 'successful_commands'
        except TypeError:pass# TypeError: unsupported operand type(s) for /=: 'float' and 'str'
        # rp.ptoc()
        return None if origin else out
    c=candidates
    o=origin
    import rp
    # print('co',len(c),len(o))
    import re
    compiled=re.compile('.*'+'.*'.join(re.escape(x) for x in origin.lower())+'.*')#Using regex to weed out bad candidates sped this function up by an order of magnitude when we had 4000 candidates
    c=candidates=[c for c in candidates if compiled.fullmatch(c.lower())]#Speed things up and weed out bad candidates
    out=sorted([x[1] for x in sorted(x for x in zip([match(o,x)for x in c],c) if x[0] != None) if x[1] != 'mro'],key=lambda x:x==x.startswith('_')+(x.startswith('__') and x.endswith('__')))
    # rp.ptoc()
    _ryan_completion_matches_cache[h]=out
    return out
 



from typing import Any

COLLECTION_ACTIONS: Any
MEMBER_ACTIONS: Any

def strip_slashes(name): ...

class SubMapperParent:
    def submapper(self, **kargs): ...
    def collection(self, collection_name, resource_name, path_prefix=..., member_prefix=..., controller=...,
                   collection_actions=..., member_actions=..., member_options=..., **kwargs): ...

class SubMapper(SubMapperParent):
    kwargs: Any
    obj: Any
    collection_name: Any
    member: Any
    resource_name: Any
    formatted: Any
    def __init__(self, obj, resource_name=..., collection_name=..., actions=..., formatted=..., **kwargs) -> None: ...
    def connect(self, *args, **kwargs): ...
    def link(self, rel=..., name=..., action=..., method=..., formatted=..., **kwargs): ...
    def new(self, **kwargs): ...
    def edit(self, **kwargs): ...
    def action(self, name=..., action=..., method=..., formatted=..., **kwargs): ...
    def index(self, name=..., **kwargs): ...
    def show(self, name=..., **kwargs): ...
    def create(self, **kwargs): ...
    def update(self, **kwargs): ...
    def delete(self, **kwargs): ...
    def add_actions(self, actions): ...
    def __enter__(self): ...
    def __exit__(self, type, value, tb): ...

class Mapper(SubMapperParent):
    matchlist: Any
    maxkeys: Any
    minkeys: Any
    urlcache: Any
    prefix: Any
    req_data: Any
    directory: Any
    always_scan: Any
    controller_scan: Any
    debug: Any
    append_slash: Any
    sub_domains: Any
    sub_domains_ignore: Any
    domain_match: Any
    explicit: Any
    encoding: Any
    decode_errors: Any
    hardcode_names: Any
    minimization: Any
    create_regs_lock: Any
    def __init__(self, controller_scan=..., directory=..., always_scan=..., register=..., explicit=...) -> None: ...
    environ: Any
    def extend(self, routes, path_prefix=...): ...
    def make_route(self, *args, **kargs): ...
    def connect(self, *args, **kargs): ...
    def create_regs(self, *args, **kwargs): ...
    def match(self, url=..., environ=...): ...
    def routematch(self, url=..., environ=...): ...
    obj: Any
    def generate(self, *args, **kargs): ...
    def resource(self, member_name, collection_name, **kwargs): ...
    def redirect(self, match_path, destination_path, *args, **kwargs): ...



from typing import Any, Optional

digest_size: Any

class HMAC:
    digest_size: Any
    digestmod: Any
    outer: Any
    inner: Any
    def __init__(self, key, msg: Optional[Any] = ..., digestmod: Optional[Any] = ...) -> None: ...
    def update(self, msg): ...
    def copy(self): ...
    def digest(self): ...
    def hexdigest(self): ...

def new(key, msg: Optional[Any] = ..., digestmod: Optional[Any] = ...): ...



from dataclasses import dataclass, field

from ..components.autocomplete_popup.autocomplete_popup_keymapper import \
    AutoCompletePopUpKeyMapper
from ..components.json_browser.json_browser_keymapper import \
    JSONBrowserKeyMapper
from ..components.query_bar.query_bar_keymapper import QueryBarKeyMapper


class InputFilter:

    def __init__(self, global_command_key):
        self.global_command_key = global_command_key
        self.wait_for_second_stroke = False

    def filter(self, keys, raw):
        if self.wait_for_second_stroke:
            self.wait_for_second_stroke = False
            combined_keys = [self.global_command_key + " " + keys[0]]
            combined_keys.extend(self.combine(keys[1:]))
            return combined_keys

        combined_keys = self.combine(keys)

        if combined_keys[-1] == self.global_command_key:
            self.wait_for_second_stroke = True
            return combined_keys[:-1]

        return combined_keys

    def combine(self, keys):
        """
        Search and combine global_command_key with the next key
        """
        combined_keys = []

        index = 0
        while index < len(keys):
            key = keys[index]

            if index == len(keys) - 1 or key != self.global_command_key:
                combined_keys.append(key)
                index += 1
                continue

            combined_keys.append(
                self.global_command_key + " " + keys[index + 1]
            )
            index += 2

        return combined_keys


@dataclass(frozen=True)
class KeyMapper:

    exit: str = "q"

    global_command_key: str = None
    input_filter: InputFilter = field(init=False)

    json_browser: JSONBrowserKeyMapper = JSONBrowserKeyMapper()
    query_bar: QueryBarKeyMapper = QueryBarKeyMapper()
    autocomplete_popup: AutoCompletePopUpKeyMapper = \
        AutoCompletePopUpKeyMapper()

    def __post_init__(self):
        object.__setattr__(
            self, "input_filter", InputFilter(self.global_command_key)
        )



from __future__ import unicode_literals, absolute_import

from rp.prompt_toolkit.enums import DEFAULT_BUFFER
from rp.prompt_toolkit.filters import Condition
from rp.prompt_toolkit.layout.controls import TokenListControl

from rp.rp_ptpython.prompt_style import PromptStyle

from pygments.token import Token
from pygments.lexers import PythonLexer

import linecache
import os

__all__ = (
    'PdbPromptStyle',
    'CallStack',
    'format_stack_entry',
)

class PdbPromptStyle(PromptStyle):
    """
    Pdb prompt.

    Show "(pdb)" when we have a pdb command or '>>>' when the user types a
    Python command.
    """
    def __init__(self, pdb_commands):
        self.pdb_commands = pdb_commands

    def in_tokens(self, cli):
        b = cli.buffers[DEFAULT_BUFFER]

        command = b.document.text.lstrip()
        if command:
            command = command.split()[0]

        if any(c.startswith(command) for c in self.pdb_commands):
            return [(Token.Prompt, '(pdb) ')]
        else:
            return [(Token.Prompt, '  >>> ')]

    def in2_tokens(self, cli, width):
        return [
            (Token.Prompt, '  ... ')
        ]

    def out_tokens(self, cli):
        return []  # Not used.


python_lexer = PythonLexer(
    stripnl=False,
    stripall=False,
    ensurenl=False)


class CallStack(TokenListControl):
    def __init__(self, pdb_ref):
        def get_tokens(cli):
            """
            See bdb.py, format_stack_entry.
            """
            pdb = pdb_ref()
            result = []

            for i, (frame, lineno) in enumerate(pdb.stack):
                is_selected = i == pdb.callstack_selected_frame
                has_focus = is_selected and pdb.callstack_focussed

                result.extend(format_stack_entry(pdb, frame, lineno, has_focus))

                # Focus cursor.
                if is_selected:
                    result.append((Token.SetCursorPosition, ' '))

                result.append((Token, '\n'))

            return result

        super(CallStack, self).__init__(
            get_tokens, has_focus=Condition(lambda cli: pdb_ref().callstack_focussed))


def format_stack_entry(pdb, frame, lineno, has_focus=False):
    result = []

    if frame is pdb.curframe:
        result.append((Token.CurrentLine, '->'))
        result.append((Token, ' '))
    else:
        result.append((Token, '   '))

    filename = pdb.canonic(frame.f_code.co_filename)

    # Filename/lineno
    token = Token.Name.Selected if has_focus else Token.Name
    result.append((token, os.path.basename(filename)))
    result.append((Token.Number, '(%r)' % lineno))

    # co_name
    if frame.f_code.co_name:
        result.append((Token.Name, frame.f_code.co_name))
    else:
        result.append((Token.Name, '<lambda>'))

    # Args.
    if '__args__' in frame.f_locals:
        args = frame.f_locals['__args__']
        result.append((Token.Name, repr(args)))
    else:
        result.append((Token.Punctuation, '()'))

    # Return value.
    if '__return__' in frame.f_locals:
        rv = frame.f_locals['__return__']
        result.append((Token.Operator, '->'))
        result.append((Token, repr(rv)))

    result.append((Token, '\n'))
    if has_focus:
        result.append((Token.SelectedFrame, ''))
    result.append((Token, '     '))

    line = linecache.getline(filename, lineno, frame.f_globals).strip()
    result.extend(python_lexer.get_tokens(line))

    return result
